#include "AABB.h"
#include "udebug.h"

//
// costructor
//
AABB3::AABB3( float xMin, float xMax, float yMin, float yMax, float zMin, float zMax )
: m_minX( xMin ), m_maxX( xMax ), m_minY( yMin ), m_maxY( yMax ), m_minZ( zMin ), m_maxZ( zMax )
{

}

//
// set 
//
void AABB3::setAABB( float minX, float maxX, float minY, float maxY, float minZ, float maxZ )
{
    m_minX = minX;
    m_maxX = maxX;
    m_minY = minY;
    m_maxY = maxY;
    m_minZ = minZ;
    m_maxZ = maxZ;
}

//
// check if this object intersects with arg
//
//   ġ ġ  Ѵ.
//
bool AABB3::intersect( AABB3 arg )
{
    // ġ ʴ  test
    // x-axis
    if( ( m_maxX < arg.m_minX ) || ( m_minX > arg.m_maxX ) )
    {
        return false;
    }
    // y-axis
    if( ( m_maxY < arg.m_minY ) || ( m_minY > arg.m_maxY ) )
    {
        return false;
    }
    // z-axis
    if( ( m_maxZ < arg.m_minZ ) || ( m_minZ > arg.m_maxZ ) )
    {
        return false;
    }
    return true;

}

#define intersectT( x, y, z ) \
    D3DXIntersectTri( &ev[x], &ev[y], &ev[z], vPickOrg, vPickDir, &u, &v, &t );

//    intersectTriangle( ev[x], ev[y], ev[z], *vPickOrg, *vPickDir, &t, &u, &v )

//#define INTERSECT_DEBUG

//
// check intersection with line
//
bool AABB3::intersect( const D3DXVECTOR3 *vPickOrg, const D3DXVECTOR3 *vPickDir )
{
    // get edge vertex
    D3DXVECTOR3 ev[8];
    getVertices( ev );

    // check all plains
    float t, u, v;
    BOOL ret;

    // bottom
    ret = intersectT( 0, 1, 3 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("bottom 1");
#endif
        return true ;
    }
    ret = intersectT( 1, 2, 3 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("bottom 2");
#endif
        return true;
    }
    // top
    ret = intersectT( 4, 5, 7 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("top 1");
#endif
        return true;
    }
    ret = intersectT( 5, 6, 7 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("top 2");
#endif
        return true;
    }

    // left
    ret = intersectT( 0, 1, 4 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("left 1");
#endif
        return true;
    }
    ret = intersectT( 0, 4, 5 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("left 2");
#endif
        return true;
    }

    // right
    ret = intersectT( 2, 3, 7 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("right 1");
#endif
        return true;
    }
    ret = intersectT( 2, 7, 6 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("right 2");
#endif
        return true;
    }
    
    // front
    ret = intersectT( 0, 4, 3 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("front 1");
#endif
        return true;
    }
    ret = intersectT( 3, 4, 7 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("fornt 2");
#endif
        return true;
    }

    // back
    ret = intersectT( 1, 5, 2 );
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("back 1");
#endif
        return true;
    }
    ret = intersectT( 2, 5, 6 );
    
    if( ret == TRUE ) 
    {
#ifdef INTERSECT_DEBUG
        UPRINT("back 2");
#endif
        return true;
    }
#ifdef INTERSECT_DEBUG
    UPRINT("no intersect");
#endif
    return false;
}

//
// check if this object contains arg
//
bool AABB3::contains( AABB3 arg )
{
    return ( m_maxX >= arg.m_maxX ) && ( m_minX <= arg.m_minX )
        && ( m_maxY >= arg.m_maxY ) && ( m_minY <= arg.m_minY )
        && ( m_maxZ >= arg.m_maxZ ) && ( m_minZ <= arg.m_minZ );
}

//
// initialize AABB
//
void AABB3::init( float x, float y, float z )
{
    m_minX = m_maxX = x;
    m_minY = m_maxY = y;
    m_minZ = m_maxZ = z;
}

void AABB3::init( D3DXVECTOR3 v )
{
    m_minX = m_maxX = v.x;
    m_minY = m_maxY = v.y;
    m_minZ = m_maxZ = v.z;
}

//
// object   ߰ϰ AABB Ѵ.
//
void AABB3::addPoint( float x, float y, float z )
{
    if( m_minX > x ) {
        m_minX = x;
    }
    if( m_maxX < x ) {
        m_maxX = x;
    }
    if( m_minY > y ) {
        m_minY = y;
    }
    if( m_maxY < y ) {
        m_maxY = y;
    }
    if( m_minZ < z ) {
        m_minZ = z;
    }
    if( m_maxZ > z ) {
        m_maxZ = z;
    }
}

void AABB3::addPoint( D3DXVECTOR3 v )
{
    if( m_minX > v.x ) {
        m_minX = v.x;
    }
    if( m_maxX < v.x ) {
        m_maxX = v.x;
    }
    if( m_minY > v.y ) {
        m_minY = v.y;
    }
    if( m_maxY < v.y ) {
        m_maxY = v.y;
    }
    if( m_minZ > v.z ) {
        m_minZ = v.z;
    }
    if( m_maxZ < v.z ) {
        m_maxZ = v.z;
    }
}

//
// get the corners of the box
//
void AABB3::getVertices( D3DXVECTOR3 vCorner[8] )
{
    vCorner[0].x = m_minX;
    vCorner[0].y = m_minY;
    vCorner[0].z = m_minZ;

    vCorner[1].x = m_minX;
    vCorner[1].y = m_minY;
    vCorner[1].z = m_maxZ;

    vCorner[2].x = m_maxX;
    vCorner[2].y = m_minY;
    vCorner[2].z = m_maxZ;

    vCorner[3].x = m_maxX;
    vCorner[3].y = m_minY;
    vCorner[3].z = m_minZ;

    vCorner[4].x = m_minX;
    vCorner[4].y = m_maxY;
    vCorner[4].z = m_minZ;

    vCorner[5].x = m_minX;
    vCorner[5].y = m_maxY;
    vCorner[5].z = m_maxZ;

    vCorner[6].x = m_maxX;
    vCorner[6].y = m_maxY;
    vCorner[6].z = m_maxZ;

    vCorner[7].x = m_maxX;
    vCorner[7].y = m_maxY;
    vCorner[7].z = m_minZ;
}

//
// get the corners of the box
//
void AABB3::getVertices( D3DXVECTOR4 vCorner[8] )
{
    vCorner[0].x = m_minX;
    vCorner[0].y = m_minY;
    vCorner[0].z = m_minZ;
    vCorner[0].w = 1;

    vCorner[1].x = m_minX;
    vCorner[1].y = m_minY;
    vCorner[1].z = m_maxZ;
    vCorner[1].w = 1;

    vCorner[2].x = m_minX;
    vCorner[2].y = m_maxY;
    vCorner[2].z = m_minZ;
    vCorner[2].w = 1;

    vCorner[3].x = m_minX;
    vCorner[3].y = m_maxY;
    vCorner[3].z = m_maxZ;
    vCorner[3].w = 1;

    vCorner[4].x = m_maxX;
    vCorner[4].y = m_minY;
    vCorner[4].z = m_minZ;
    vCorner[4].w = 1;

    vCorner[5].x = m_maxX;
    vCorner[5].y = m_minY;
    vCorner[5].z = m_maxZ;
    vCorner[5].w = 1;

    vCorner[6].x = m_maxX;
    vCorner[6].y = m_maxY;
    vCorner[6].z = m_minZ;
    vCorner[6].w = 1;

    vCorner[7].x = m_maxX;
    vCorner[7].y = m_maxY;
    vCorner[7].z = m_maxZ;
    vCorner[7].w = 1;
}

//
// operator overloading "inequality(!=)"
//
bool AABB3::operator!=( const AABB3& arg )
{
    return (m_minX != arg.m_minX)
        || (m_maxX != arg.m_maxX)
        || (m_minY != arg.m_minY)
        || (m_maxY != arg.m_maxY)
        || (m_minZ != arg.m_minZ)
        || (m_maxZ != arg.m_maxZ);
}

//
// operator overloading "assignment( = )"
//
void AABB3::operator =( const AABB3 &in )
{
    m_minX = in.m_minX;
    m_maxX = in.m_maxX;
    m_minY = in.m_minY;
    m_maxY = in.m_maxY;
    m_minZ = in.m_minZ;
    m_maxZ = in.m_maxZ;
}