#include "MapEditor.h"
#include "TmpTri.h"
#include "udebug.h"
#include "util.h"

char strFrameRate[1024] = "";

//
// ʱȭ
//
HRESULT MapEditor::init( HWND hWnd )
{
    // create the D3D object
    if( NULL == (m_pD3D = Direct3DCreate9( D3D_SDK_VERSION )) )
    {
        return E_FAIL;
    }

    // set up the structure used to create the D3DDevice
    ZeroMemory( &m_d3dpp, sizeof( m_d3dpp ) );
    m_d3dpp.Windowed = true;
    m_d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    m_d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
    m_d3dpp.EnableAutoDepthStencil = TRUE;
    m_d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

    // Create the D3DDevice
    if( FAILED( m_pD3D->CreateDevice( D3DADAPTER_DEFAULT,
                                      D3DDEVTYPE_HAL,
                                      hWnd,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                      &m_d3dpp,
                                      &m_pd3dDevice ) ) )
    {
        return E_FAIL;
    }

    RECT logRegion;
    GetClientRect( hWnd, &logRegion );
    // initialize log modules
    initLog( m_pd3dDevice, logRegion, 10 );

    // Turn off culling, so we see the front and back of the triangle
    //m_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

    // Turn on the zbuffer
    m_pd3dDevice->SetRenderState( D3DRS_ZENABLE, TRUE );

    // To enable material alpha
//    m_pd3dDevice->SetRenderState( D3DRS_DIFFUSEMATERIALSOURCE, D3DMCS_MATERIAL );

//    m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );

    // Set the source blend state.
    m_pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCCOLOR);

    // Set the destination blend state.
    m_pd3dDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCCOLOR);

    // initialize brush
    m_brushSize = BRUSH_1;
    m_brushShape = SQUARE_BRUSH;
    m_cmd = CMD_NONE;

    m_prevCursorIdx = -1;

    setupLights();

    // initialize timer
    m_timer.init();

    isMapAvailable = false;

    return S_OK;
}

//
// ڿ 
//
void MapEditor::release()
{   
    // release log module
    releaseLog();

    if( m_pd3dDevice != NULL )
    {
        m_pd3dDevice->Release();
    }
    if( m_pD3D != NULL )
    {
        m_pD3D->Release();
    }

    isMapAvailable = false;
}

//
// render a scene
//
void MapEditor::render()
{
    // clear the backbuffer and the zbuffer
    m_pd3dDevice->Clear( 0,
                         NULL,
                         D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                         D3DCOLOR_XRGB( 0,0,0 ),
                         1.0f,
                         0 );

    // set FPS
    int fps = m_timer.getFPS();

    if( fps )
    {
        sprintf( strFrameRate,"FPS : %d", fps );
        Log::setFPSStr(strFrameRate);
    }

    // set render mode
    m_pd3dDevice->SetRenderState( D3DRS_FILLMODE, m_viewState );

    // set camera
    m_camera.setCamera( m_pd3dDevice );

    // Begin the scene
    if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
    {
        // print log
        Log::render( m_pd3dDevice );

        m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );

        // render a map
        m_map.render( m_camera.getFrustum() );

        m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );

        // draw cursor
        m_map.drawCursor( m_pd3dDevice, m_cursorIdx, m_brushShape, m_brushSize );

        // end the scene
        m_pd3dDevice->EndScene();
    }

    // Present the backbuffer contents to the display
    m_pd3dDevice->Present( NULL, NULL, NULL, NULL );
}

//
//  
//
void MapEditor::setupLights()
{
    // set up a white, directional light
    D3DXVECTOR3 vecDir;
    D3DLIGHT9 light;
    ZeroMemory( &light, sizeof( D3DLIGHT9 ) );
    light.Type = D3DLIGHT_DIRECTIONAL;
    D3DXCOLOR color( 1.0f, 1.0f, 1.0f, 1.0f );
    light.Ambient = color * 0.6f;
    light.Diffuse = color;
    light.Specular = color  * 0.6f;
    vecDir = D3DXVECTOR3( cosf( D3DX_PI/4),
                          -sinf( D3DX_PI/4),
                          0 );
    D3DXVec3Normalize( (D3DXVECTOR3*)&light.Direction, &vecDir );
    light.Range = 10.0f;
    m_pd3dDevice->SetLight( 0, &light );
    m_pd3dDevice->LightEnable( 0, TRUE );
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE ); 

}

//
// create a new map
//
void MapEditor::createNewMap( int row, int col )
{
    m_map.create( m_pd3dDevice, row, col );
    m_camera.init( D3DXVECTOR3( (float)(row-1)/2, 0, (float)(col-1)/2 ) );

    isMapAvailable = true;
}

//
// change the view mode to next view mode
// solid <-> wire frame
//
void MapEditor::nextViewMode()
{
    m_viewState = (m_viewState == D3DFILL_SOLID)?D3DFILL_WIREFRAME:D3DFILL_SOLID;
}

//
// get current command
//
char * MapEditor::getCmd()
{
    static char *cmdStr[3];
    cmdStr[0] = "None";
    cmdStr[1] = "Raise";
    cmdStr[2] = "Lower";

    return cmdStr[ m_cmd ];
}

//
// get brush size
//
char * MapEditor::getBrushSize()
{
    static char *brushStr[9];
    brushStr[1] = "1";
    brushStr[2] = "2";
    brushStr[3] = "3";
    brushStr[5] = "5";
    brushStr[8] = "8";

    return brushStr[ m_brushSize ];
}

//
// get brush shape
//
char * MapEditor::getBrushShape()
{
    static char *brushShapeStr[2];
    brushShapeStr[0] = "Circle";
    brushShapeStr[1] = "Square";

    return brushShapeStr[ m_brushShape ];
}

//
// change a command
//
void MapEditor::changeCmd( int cmd )
{
    m_cmd = cmd;
}

//
// change the size of a brush
//
void MapEditor::changeBrushSize( int size )
{
    m_brushSize = size;
    // change previous cursor index
    // to remake the area of cursor
    m_prevCursorIdx = -1;
}

//
// change the shape of a brush
//
void MapEditor::changeBrushShape( int shape )
{
    m_brushShape = shape;
    // change previous cursor index
    // to remake the area of cursor
    m_prevCursorIdx = -1;
}

//
// do command
//
// assume updateCursorPos is already called
//
void MapEditor::doCmd( int x, int y )
{
    if( m_cmd == CMD_NONE || m_cursorIdx == -1 )
    {
        return;
    }

    float delta;

    switch( m_cmd )
    {
    case CMD_RAISE:
        delta = 0.05f;
        break;
    case CMD_LOWER:
        delta = -0.05f;
        break;
    }

    m_map.changeHeight( delta );
}

//#define DEBUG_PICK

//
// update cursor position
//
void MapEditor::updateCursorPos( int x, int y )
{
    if( m_map.getRow() != 0 ) {
        D3DVIEWPORT9 vp;
        m_pd3dDevice->GetViewport( &vp );

        D3DXVECTOR3 vRayDirNDC((float)x,(float)y,0.0f);
        D3DXVECTOR3 vRayDirB((float)x,(float)y,1.0f);
        D3DXVECTOR3 vNPt;
        D3DXVECTOR3 vNPtB;

        D3DXVec3Unproject(&vNPt,  &vRayDirNDC, &vp, &(m_camera.getMatProj()), &(m_camera.getMatView()), NULL);
        D3DXVec3Unproject(&vNPtB, &vRayDirB,   &vp, &(m_camera.getMatProj()), &(m_camera.getMatView()), NULL);

        vNPtB = vNPtB-vNPt;

        D3DXVECTOR3 *pEye = &vNPt;
        D3DXVECTOR3 *pRayDir = &vNPtB;

        m_cursorIdx = m_map.findCursorVertex( pRayDir, pEye );

        if( m_cursorIdx != m_prevCursorIdx )
        {
            m_map.makeCursorArea( m_cursorIdx, m_brushShape, m_brushSize );
            m_prevCursorIdx = m_cursorIdx;
        }

#ifdef DEBUG_PICK
        //UPRINT( "vp(%f, %f, %f, %f, %f, %f)", vp.X, vp.Y, vp.Width, vp.Height, vp.MinZ, vp.MaxZ );
        UPRINT("(%d, %d)", x, y );
        D3DXVECTOR3 vPickRayDir, vPickRayOrig;
        vPickRayDir = *pRayDir;
        vPickRayOrig = *pEye;
        UPRINT( "dir(%f,%f,%f), org(%f,%f,%f)", vPickRayDir.x, vPickRayDir.y, vPickRayDir.z, vPickRayOrig.x, vPickRayOrig.y, vPickRayOrig.z );
        UPRINT( "m_cursorIdx=%d, cursor(%d,%d)", m_cursorIdx, m_cursorIdx%m_map.m_row, m_cursorIdx/m_map.m_row );
        D3DXPLANE xyPlane( 0, 1, 0, 0 );
        D3DXVECTOR3 endP = *pEye + (*pRayDir)*10;
        D3DXVECTOR3 out;
        D3DXPlaneIntersectLine( &out, &xyPlane, pEye, &endP );
        UPRINT( "out( %f, %f, %f )", out.x, out.y, out.z );
#endif
    }
}

//
// save to a file
// 
void MapEditor::save( const char *filename )
{
    m_map.save( filename );
}

//
// load from a file
//
void MapEditor::load( const char *filename )
{
    m_map.load( m_pd3dDevice, filename );
    m_camera.init( D3DXVECTOR3( (float)(m_map.getRow()-1)/2, 0, (float)(m_map.getCol()-1)/2 ) );

    isMapAvailable = true;
}

//
// get a frustum on a x-z plane
//
void MapEditor::getFrustumOnXZ( D3DXVECTOR3 *edgeVertices )
{
    m_camera.getFrustumOnXZ( edgeVertices );
}

//
// move a camera
//
void MapEditor::moveCamera( const D3DXVECTOR3 *lookat )
{
    m_camera.move( *lookat );
}

void MapEditor::moveCamera( const int direction )
{
    static float delta = 0.2f;

    switch( direction )
    {
    case LEFT :
        m_camera.moveLeft( delta );
        break;
    case RIGHT :
        m_camera.moveRight( delta );
        break;
    case FRONT :
        m_camera.moveFront( delta );
        break;
    case BACK :
        m_camera.moveBack( delta );
        break;
    }
}

//
// zoom a camera
//
void MapEditor::zoomCamera( const float delta )
{
    m_camera.zoom( delta );
}

//
// rotate a camera
//
void MapEditor::rotateCamera( const float x, const float y)
{
    m_camera.rotate( x*2.0f, y*2.0f );
}

//
// reset window size
//
void MapEditor::resetWindowSize( int width, int height )
{
    m_d3dpp.BackBufferWidth = width;
    m_d3dpp.BackBufferHeight = height;
    m_pd3dDevice->Reset( &m_d3dpp );
}

//
// get a size of a map
//
void MapEditor::getMapSize( int *row, int *col )
{
    *row = m_map.getRow();
    *col = m_map.getCol();
}