//
// main application class
//

#ifndef __MAPEDITOR_H__
#define __MAPEDITOR_H__

#include <d3dx9.h>
#include "UMap.h" 
#include "UCamera.h"
#include "UTimer.h"

class MapEditor
{
private :
    // d3d  
    LPDIRECT3D9 m_pD3D;
    LPDIRECT3DDEVICE9 m_pd3dDevice;

    D3DPRESENT_PARAMETERS m_d3dpp;

    // map
    UMap m_map;

    // camera
    UCamera m_camera;

    // view state
    DWORD m_viewState;

    // picked cursor index
    int m_cursorIdx;

    // previous cursor index
    int m_prevCursorIdx;

    // brush size(1-2-3-5-8)
    int m_brushSize;

    // timer
    UTimer m_timer;

public :
    // a flag notify if a map is available
    bool isMapAvailable; 

    // constants defines brush size
    const static int BRUSH_1 = 1;
    const static int BRUSH_2 = 2;
    const static int BRUSH_3 = 3;
    const static int BRUSH_5 = 5;
    const static int BRUSH_8 = 8;

private :
    // brush shape
    int m_brushShape;

public :
    // contants defines brush shapes
    const static int CIRCLE_BRUSH = 0;
    const static int SQUARE_BRUSH = 1;

private :
    // command
    int m_cmd;

public :
    // constants defines commands
    const static int CMD_NONE  = 0;
    const static int CMD_RAISE = 1;
    const static int CMD_LOWER = 2;

    // constants for moving a camera
    const static int LEFT = 0;
    const static int RIGHT = 1;
    const static int FRONT = 2;
    const static int BACK = 3;

private :
    // setup lights for show the colors of objects
    void setupLights();

public :
    // ʱȭ
    HRESULT init( HWND hWnd );
    // release
    void release();
    // render
    void render();
    // create new map
    void createNewMap( int row, int col );
    // change to next view mode
    void nextViewMode();

    // do command
    void doCmd( int x, int y);
    // update cursor position
    void updateCursorPos( int x, int  y );

    // a operation relate to tool
    // change brush size
    void changeBrushSize( int size );
    // change brush shape
    void changeBrushShape( int shape );
    // change command
    void changeCmd( int cmd );

    // move a camera
    void moveCamera( const D3DXVECTOR3 *lookat );
    void moveCamera( const int direction );
    // zoom a camera
    void zoomCamera( const float delta );
    // rotate a camera
    void rotateCamera( const float x, const float y);

    // get current command
    char *getCmd();
    // get current brush size
    char *getBrushSize();
    // get the shape of a current brush
    char *getBrushShape();

    // file operation
    void save( const char * filename );
    void load( const char * filename );

    // get a frustum on a x-z plane
    void getFrustumOnXZ( D3DXVECTOR3 *edgeVertices );

    // reset window size
    void resetWindowSize( int width, int height );

    // get a size of a map
    void getMapSize( int *row, int *col );
};

#endif