#include "MapPatch.h"
#include "UUtil.h"
#include "udebug.h"

#ifdef EDITOR_BUILD
//
// grid ׸  Ǵ world matrix
//
D3DXMATRIX MapPatch::m_matLine
= D3DXMATRIX( 1.0f, 0,    0,    0,
              0,    1.0f, 0,    0,
              0,    0,    1.0f, 0,
              0,    0.01f, 0,    1.0f );
//
// grid ׸  Ǵ material
//
D3DMATERIAL9 MapPatch::m_mtrlLine;
#endif

//
// constructor
//
MapPatch::MapPatch( LPDIRECT3DDEVICE9 pd3dDevice, int left, int right, int bottom, int top, MapVertex* pMapVertices, int rowOffset )
: m_pd3dDevice( pd3dDevice ), m_numVertices( (right-left+1)*(top-bottom+1) ), m_primCount( (right-left)*(top-bottom)*2 )
#ifdef EDITOR_BUILD
, m_primCountLine( (right-left)*(top-bottom)*2 + (right-left) + (top-bottom ) )
#endif
{
    HRESULT hr;
    // vertex buffer 
    if( FAILED(hr = m_pd3dDevice->CreateVertexBuffer(
        m_numVertices*sizeof(MapVertex),
        D3DUSAGE_WRITEONLY,
        D3DFVF_MAPVERTEX,
        D3DPOOL_DEFAULT,
        &m_pVB,
        0)) )
    {
        switch( hr )
        {
        case D3DERR_INVALIDCALL :
            PRINTSCR( "Fail : create vertex buffer( INVALIDCALL )" );
            break;
        case D3DERR_OUTOFVIDEOMEMORY :
            PRINTSCR( "Fail : create vertex buffer( OUTOFVIDEOMEMORY )" );
            break;
        case E_OUTOFMEMORY :
            PRINTSCR( "Fail : create vertex buffer( E_OUTOFMEMORY )" );
            break;
        }
    }

    // index buffer 
    if( FAILED(hr = m_pd3dDevice->CreateIndexBuffer(
        m_primCount * 3 * sizeof(WORD),
        D3DUSAGE_WRITEONLY,
        D3DFMT_INDEX16,
        D3DPOOL_DEFAULT,
        &m_pIB,
        0) ) )
    {
        PRINTSCR( "Fail : create index buffer" );
    }

#ifdef EDITOR_BUILD
    // Grid  index buffer 
    if( FAILED(hr = m_pd3dDevice->CreateIndexBuffer(
        m_numVertices * 2 * sizeof(WORD),
        D3DUSAGE_WRITEONLY,
        D3DFMT_INDEX16,
        D3DPOOL_DEFAULT,
        &m_pGridIB,
        0) ) )
    {
        PRINTSCR( "Fail : create grid index buffer" );
    }
#endif

    // vertex  
    MapVertex* pVertex;
    if( FAILED(hr = m_pVB->Lock( 0, 0, (void**)&pVertex, 0) ) )
    {
        PRINTSCR( "Fail : lock vertex buffer" );
    }
    int index;
    int count=0;
    for( int i=bottom; i<=top; i++ )
    {
        for( int j=left; j<=right; j++ )
        {
            index = i*rowOffset + j;
            pVertex[count++] = pMapVertices[index];
        }
    }
    m_pVB->Unlock();

    WORD *pIndices;
    // index  
    if( FAILED(hr = m_pIB->Lock( 0, 0, (void **)&pIndices, 0) ) )
    {
        PRINTSCR( "Fail : lock index buffer" );
    }
    int row = right-left;
    int height = top-bottom;
    int rowSize = row+1;
    count = 0;
    for( int i=0; i<height; i++)
    {
        for( int j=0; j<row; j++ )
        {
            index = i*rowSize + j;
            pIndices[count++] = index;
            pIndices[count++] = index + rowSize;
            pIndices[count++] = index + 1;
            pIndices[count++] = index + 1;
            pIndices[count++] = index + rowSize;
            pIndices[count++] = index + rowSize + 1;
        }
    }
    m_pIB->Unlock();

#ifdef EDITOR_BUILD
    // index  
    if( FAILED(hr = m_pGridIB->Lock( 0, 0, (void **)&pIndices, 0) ) )
    {
        PRINTSCR( "Fail : lock index buffer" );
    }
    count = 0;
    for( int i=0; i<=height; i++)
    {
        for( int j=0; j<=row; j++ )
        {
            index = i*rowSize + j;
            // up line
            if( i != height ) {
                pIndices[count ++] = index; // current index
                pIndices[count ++] = index + rowSize;  // upper index
            }
            // right line
            if( j != row ) 
            {
                pIndices[count ++] = index; // current index
                pIndices[count ++] = index + 1;
            }            
        }
    }
    m_pGridIB->Unlock();

    // initialize the property of material
    ZeroMemory( &m_mtrlLine, sizeof( D3DMATERIAL9 ) );
    m_mtrlLine.Diffuse.r = m_mtrlLine.Ambient.r = 1.0f;
    m_mtrlLine.Diffuse.g = m_mtrlLine.Ambient.g = 1.0f;
    m_mtrlLine.Diffuse.b = m_mtrlLine.Ambient.b = 1.0f;
    m_mtrlLine.Diffuse.a = m_mtrlLine.Ambient.a = 0.0f;    
#endif
}

#ifdef EDITOR_BUILD
//
// patch Ѵ.
//
void MapPatch::update( int left, int right, int bottom, int top, MapVertex *pMapVertices, int rowOffset )
{
    // vertex  
    MapVertex* pVertex;
    if( FAILED( m_pVB->Lock( 0, 0, (void**)&pVertex, 0) ) )
    {
        PRINTSCR( "Fail : lock vertex buffer" );
    }
    int index;
    int count=0;
    for( int i=bottom; i<=top; i++ )
    {
        for( int j=left; j<=right; j++ )
        {
            index = i*rowOffset + j;
            pVertex[count++] = pMapVertices[index];
        }
    }
    m_pVB->Unlock();
}
#endif

//
// ڿ ȯѴ
//
void MapPatch::release()
{
    m_pd3dDevice = 0;
    ReleaseCom( m_pVB );
    ReleaseCom( m_pIB );
#ifdef EDITOR_BUILD
    ReleaseCom( m_pGridIB );
#endif
}

//
// map ׸.
//
void MapPatch::render()
{
    m_pd3dDevice->SetStreamSource( 0, m_pVB, 0, sizeof( MapVertex ) );
    m_pd3dDevice->SetIndices( m_pIB );
    m_pd3dDevice->DrawIndexedPrimitive( 
        D3DPT_TRIANGLELIST,
        0,
        0,
        m_numVertices,
        0,
        m_primCount);

#ifdef EDITOR_BUILD
    m_pd3dDevice->SetMaterial( &m_mtrlLine );
    m_pd3dDevice->SetIndices( m_pGridIB );
    m_pd3dDevice->DrawIndexedPrimitive(
        D3DPT_LINELIST ,       
        0,                     
        0,
        m_numVertices,
        0,
        m_primCountLine );
#endif
}
