#ifndef __QTREE_H__
#define __QTREE_H__

#include <vector>
#include <d3dx9.h>

#include "MapPatch.h"
#include "AABB.h"
#include "UFrustum.h"
#include "MapVertex.h"

enum {
    NW = 0,
    NE,
    SE,
    SW
};

//
// quad tree node class
//
// z+
//          top
//     
//      NW  NE 
// leftright
//      SW  SE 
//     
//         bottom
//  x+
//
class QTreeNode
{
public :
    // constructor
    QTreeNode( LPDIRECT3DDEVICE9 pd3dDevice, int depth, int left, int right, int bottom, int top, MapVertex *pMapVertices, int rowOffset, std::vector<QTreeNode *>* nodeList);

    // destructor
    ~QTreeNode()
    {
        release();
    }

    // ڿ ȯѴ.
    void release();

    // frustum cullingϸ鼭 renderѴ.
    void render( UFrustum& frustum );

    // AABB pointer Ѱش.
    AABB3* getAABB()
    {
        return m_aabb;
    }

    // node list index Ѱش.
    size_t getIdx()
    {
        return m_idx;
    }

    // 콺 Ŭ  ǥ ˷ش.
    bool findPickedPos( const D3DXVECTOR3 *vPickOrg, const D3DXVECTOR3 *vPickDir, D3DXVECTOR3 *vPickOut, float *t, const MapVertex *mapVertices, int rowOffset );

#ifdef EDITOR_BUILD
    // 콺   vertex ã´.
    void findPickedVertex( const D3DXVECTOR3 *vPickOrg, const D3DXVECTOR3 *vPickDir, int *idx, float *t, const MapVertex *mapVertices, int rowOffset);

    // update ɼ ִ AABB checkѴ.
    bool updateAABB( int x, int y, float oldYVal, float newYVal );
    // AABB  ȮѴ.
    // Patch updateѴ.
    bool commitAABBAndPatch( MapVertex* mapVertices, int rowOffset );
#endif

private :
    // 콺 Ŭ  ǥ Ѵ.
    bool findIntersectedPos( const D3DXVECTOR3 *vPickOrig, const D3DXVECTOR3 *vPickDir, D3DXVECTOR3 *vPickOut, float *tmpt, const MapVertex *mapVertex, int rowOffset );

#ifdef EDITOR_BUILD
    // 콺  ̿ ִ vertex ã´.
    void findIntersectedVertex( const D3DXVECTOR3 *vPickOrig, const D3DXVECTOR3 *vPickDir, float *tmpt, int *pickedIdx, const MapVertex *mapVertex, int rowOffset );
#endif


    // vertex  ̿ؼ aabb Ѵ.
    void createAABB( MapVertex* pMapVertices, int rowOffset );
    // child node AABB ̿ؼ aabb Ѵ.
    void createAABB( AABB3 *nwAABB, AABB3 *neAABB, AABB3 *seAABB, AABB3 *swAABB );

    // node list index
    size_t m_idx;

    // depth
    int m_depth;

    // left, right, top, bottom
    int m_left, m_right, m_top, m_bottom;

    // leaf node ƴ üũϴ flag
    bool m_isLeaf;

    // child index
    size_t m_child[4];

    // map patch
    MapPatch *m_patch;

    // Axis Aligned Bounding Box
    AABB3 *m_aabb;

#ifdef EDITOR_BUILD
    // AABB update ʿ Ȯϴ flag
    bool m_toBeUpdated;
#endif

    // node list  pointer
    std::vector<QTreeNode *>* m_pnodeList;
};

//
// quad tree class
//
class QTree
{
public :
    // constructor
    QTree()
    : m_nodeList(0), m_pMapVertices(0), m_rowOffset(0)
    {

    }

    // quad tree Ѵ.
    void create( LPDIRECT3DDEVICE9 pd3dDevice, int row, int col, MapVertex* pMapVertices );

    // ڿ Ѵ.
    void release();

    // frustum culling ϸ鼭 renderѴ.
    void render( UFrustum &frustum );

    // 콺 Ŭ  ǥ ˷ش.
    bool findPickedPos( const D3DXVECTOR3 *vPickDir, const D3DXVECTOR3 *vPickOrg, D3DXVECTOR3 *vOut );

#ifdef EDITOR_BUILD
    // 콺  ִ vertex index ã´.
    int findPickedVertex( const D3DXVECTOR3 *vPickDir, const D3DXVECTOR3 *vPickOrg );

    // update  ɼ ִ AABB üũѴ.
    void updateAABB( int x, int y, float oldYVal, float newYVal );

    // AABB Ѵ.
    void commitAABBAndPatch();
#endif

    // ְ depth
    const static int MAX_DEPTH = 5;
    //  patch ũ
    const static int MIN_PATCH_SIZE = 5;

private :
    // map vertex pointer
    MapVertex* m_pMapVertices;
    // map row ũ
    int m_rowOffset;

    // QTreeNode  list
    std::vector<QTreeNode *> m_nodeList;
};

#endif