#ifndef __UCAMERA_H__
#define __UCAMERA_H__

#include <d3dx9.h>
#include "UFrustum.h"

//
// this class model a camera
//
class UCamera
{
    // x-y direction angle
    float m_yAngle;
    // x-z direction angle
    float m_zAngle;
    // distance
    float m_distance;

    // camera's minimum distance
    static const float MIN_DISTANCE;
    // camera's maximum distance
    static const float MAX_DISTANCE;
    // camera's maximum height
    static const float MAX_YANGLE;
    // camera's minimun height
    static const float MIN_YANGLE;

    // 4 edge vertex of far plane
    static D3DXVECTOR3 f0;
    static D3DXVECTOR3 f1;
    static D3DXVECTOR3 f2;
    static D3DXVECTOR3 f3;

    static D3DXVECTOR3 vY;

    // eye position
    D3DXVECTOR3 m_eye;
    // lookat position
    D3DXVECTOR3 m_lookat;
    // up vector
    D3DXVECTOR3 m_up;

    // cash variable to calculate the view frustum
    D3DXVECTOR3 m_eyeNormal;
    float m_cosA;
    float m_sinA;
    float m_cos90A;
    float m_sin90A;
    float m_cosB;
    float m_sinB;

    // cach variable for moving camera
    D3DXVECTOR3 m_vDir;
    D3DXVECTOR3 m_vLDir;

    // Field of view in the y direction, in radians
    float m_fovy;
    // Aspect ratio, defined as view space width divided by height
    float m_aspect;
    // near view plane
    float m_near;
    // far view plane
    float m_far;

    // view matrix
    D3DXMATRIX m_matView;
    // projection matrix
    D3DXMATRIX m_matProj;

    // frustum
    UFrustum m_frustum;

    // a edge vertex of a frustum on a X-Z plane
    D3DXVECTOR3 m_edgeVertices[4];

public :

    // initialize camera
    void init( D3DXVECTOR3 lookat );

    // calculate the direction vector
    inline void calcDirVec();

    // zoom in/out
    void zoom( float fdis );

    // move the looking position to new position
    void move( D3DXVECTOR3 newPosition );

    // move a camera front
    void moveFront( const float delta );
    // move a camera back
    void moveBack( const float delta );
    // move a camera left
    void moveLeft( const float delta );
    // move a camera right
    void moveRight( const float delta );

    // rotate horizontally
    void rotateHorizontally( float angle );

    // rotate vertically
    void rotateVertically( float angle );

    // rotate horizontally and vertically
    void rotate( float zAngle, float yAngle );

    // set camera
    void setCamera( LPDIRECT3DDEVICE9 pd3dDevice );

    // calculate edge vertices of a frustum on X-Z plane
    inline void calcFrustumOnXZ();

    // calculate a polygon that intersects with a frustum on X-Z plane
    void getFrustumOnXZ( D3DXVECTOR3 *edgeVertices );

    // get
    D3DXMATRIX getMatProj()
    {
        return m_matProj;
    }
    D3DXMATRIX getMatView()
    {
        return m_matView;
    }
    UFrustum getFrustum()
    {
        return m_frustum;
    }
};

#endif