#include "UFrustum.h"

//
// generate a frustum from the view-world matrix
//
// ref : Fast Extraction of Viewing Frustum Planes from the World-View-Projection
//       Matrix
//       Gil Gribb & Klaus Hartmann
//
void UFrustum::init( D3DXMATRIX comboMatrix )
{
    // Left clipping plane
    m_planes[0].a = comboMatrix._14 + comboMatrix._11;
    m_planes[0].b = comboMatrix._24 + comboMatrix._21;
    m_planes[0].c = comboMatrix._34 + comboMatrix._31;
    m_planes[0].d = comboMatrix._44 + comboMatrix._41;
    // Right clipping plane
    m_planes[1].a = comboMatrix._14 - comboMatrix._11;
    m_planes[1].b = comboMatrix._24 - comboMatrix._21;
    m_planes[1].c = comboMatrix._34 - comboMatrix._31;
    m_planes[1].d = comboMatrix._44 - comboMatrix._41;
    // Top clipping plane
    m_planes[2].a = comboMatrix._14 - comboMatrix._12;
    m_planes[2].b = comboMatrix._24 - comboMatrix._22;
    m_planes[2].c = comboMatrix._34 - comboMatrix._32;
    m_planes[2].d = comboMatrix._44 - comboMatrix._42;
    // Bottom clipping plane
    m_planes[3].a = comboMatrix._14 + comboMatrix._12;
    m_planes[3].b = comboMatrix._24 + comboMatrix._22;
    m_planes[3].c = comboMatrix._34 + comboMatrix._32;
    m_planes[3].d = comboMatrix._44 + comboMatrix._42;
    // Near clipping plane
    m_planes[4].a = comboMatrix._13;
    m_planes[4].b = comboMatrix._23;
    m_planes[4].c = comboMatrix._33;
    m_planes[4].d = comboMatrix._43;
    // Far clipping plane
    m_planes[5].a = comboMatrix._14 - comboMatrix._13;
    m_planes[5].b = comboMatrix._24 - comboMatrix._23;
    m_planes[5].c = comboMatrix._34 - comboMatrix._33;
    m_planes[5].d = comboMatrix._44 - comboMatrix._43;
    // Normalize the plane equations, if requested
    D3DXPlaneNormalize( &m_planes[0], &m_planes[0] );
    D3DXPlaneNormalize( &m_planes[1], &m_planes[1] );
    D3DXPlaneNormalize( &m_planes[2], &m_planes[2] );
    D3DXPlaneNormalize( &m_planes[3], &m_planes[3] );
    D3DXPlaneNormalize( &m_planes[4], &m_planes[4] );
    D3DXPlaneNormalize( &m_planes[5], &m_planes[5] );
}

//
// check how a object is relate to this frustum
//
CLIPTYPE UFrustum::check( AABB3 aabb )
{
    // Frustum Culling by Dion Picco ( 01 April 2003 ) at flipcode
    D3DXVECTOR3 vCorner[8];
    int iTotalIn = 0;
    // get the corners of the box into the vCorner array
    aabb.getVertices( vCorner );

    // test all 8 corners against the 6 sides
    // if all points ar behind 1 specific plane, we are out
    // if we are in with all points, then we are fully in
    for( int p = 0; p < 6; ++p )
    {
        int iInCount = 8;
        int iPtIn = 1;
        for( int i=0; i<8; ++i )
        {
            // test this point against the planes
            if( D3DXPlaneDotCoord( &m_planes[p], &vCorner[i] ) < 0 )
            {
                iPtIn = 0;
                --iInCount;
            }
        }
        // were all the points outside of plane p?
        if( iInCount == 0 )
        {
            return FOUTSIDE;
        }
        // check if they were all on the right side of the plane
        iTotalIn += iPtIn;
    }
    // so if iTotalIn is 6, then all are inside the view
    if( iTotalIn == 6 )
    {
        return FINSIDE;
    }
    // we must be partly in then otherwise
    return FINTERSECT;

/*    D3DXVECTOR4 pos(0,0,0,1), neg(0,0,0,1);
    bool intersect = false;

    for( int i = 0; i < 6; i++ )
    {
        // x test
        if( m_planes[i].a > 0.0f )
        {
            pos.x = aabb.m_maxX;
            neg.x = aabb.m_minX;
        }
        else
        {
            pos.x = aabb.m_minX;
            neg.x = aabb.m_maxX;
        }

        // y test
        if( m_planes[i].b > 0.0f )
        {
            pos.y = aabb.m_maxY;
            neg.y = aabb.m_minY;
        }
        else
        {
            pos.y = aabb.m_minY;
            neg.y = aabb.m_maxY;
        }

        // z test
        if( m_planes[i].c > 0.0f )
        {
            pos.z = aabb.m_maxZ;
            neg.z = aabb.m_minZ;
        }
        else
        {
            pos.z = aabb.m_minZ;
            neg.z = aabb.m_maxZ;
        }

        // test negative far point
        float a = D3DXPlaneDot( &m_planes[i], &neg );
        if( a > 0 )
        {
            return FOUTSIDE;
        }

        // test positive far point
        float b = D3DXPlaneDot( &m_planes[i], &pos );

        if( b > 0 )
        {
            intersect = true;
        }
    }

    if( intersect )
    {
        return FINTERSECT;
    }

    return FINSIDE;*/
}