#include "UMap.h"
#include "UUtil.h"
#include "udebug.h"

#ifdef EDITOR_BUILD
#include <algorithm>
#include "MapEditor.h"
#endif

//
//   ϴµ Ǵ ĺ
//
char UMap::UMAP_FILE_ID []
= {
    'U', 'M', 'A', 'P',
    0x01 // version
};

#ifdef EDITOR_BUILD
//
// Ŀ  world matrix
//
// Y+  0.01 ŭ translate Ѵ.
//
D3DXMATRIX UMap::m_matCursor
= D3DXMATRIX( 1.0f, 0,    0,    0,
              0,    1.0f, 0,    0,
              0,    0,    1.0f, 0,
              0,    0.01f, 0,    1.0f );

//
// Ŀ  index
//
WORD UMap::pCursorIndex []
= {
    0,3,1,
    1,3,4,
    1,4,2,
    2,4,5,
    3,6,4,
    4,6,7,
    4,7,5,
    5,7,8
};
#endif

//
// constructor
//
UMap::UMap()
: m_row(0), m_col(0), m_pd3dDevice(0), m_pMapTexture(0), m_pMapVertices(0), m_bAvailable(false), m_pHeightMap(0)
{
    //  Ÿ material property Ѵ.
    ZeroMemory( &m_mtrlMap, sizeof( D3DMATERIAL9 ) );
    m_mtrlMap.Diffuse.r = m_mtrlMap.Ambient.r = 0.0f;
    m_mtrlMap.Diffuse.g = m_mtrlMap.Ambient.g = 0.5f;
    m_mtrlMap.Diffuse.b = m_mtrlMap.Ambient.b = 0.0f;
    m_mtrlMap.Diffuse.a = m_mtrlMap.Ambient.a = 1.0f;

#ifdef EDITOR_BUILD
    // cursor  material property Ѵ.
    ZeroMemory( &m_mtrlCursor, sizeof( D3DMATERIAL9 ) );
    m_mtrlCursor.Diffuse.r = 0.0f;
    m_mtrlCursor.Diffuse.g = 0.0f;
    m_mtrlCursor.Diffuse.b = 1.0f;
    m_mtrlCursor.Diffuse.a = 0.5f;
    m_mtrlCursor.Ambient = m_mtrlCursor.Emissive = m_mtrlCursor.Specular = m_mtrlCursor.Diffuse;
#endif
};


//
//  row-1,  col-1 ũ  0 ¥  Ѵ.
//
void UMap::create( LPDIRECT3DDEVICE9 pd3dDevice, int row, int col )
{
    // row * col ũ⸸ŭ height map Ѵ.
    float *pHeightMap = new float[ row * col ];
    // ̸ 0 Ѵ.
    ZeroMemory( pHeightMap, row*col*sizeof(float) );

    create( pd3dDevice, row, col, pHeightMap );
}

//
// map file  оδ.
//
void UMap::load( LPDIRECT3DDEVICE9 pd3dDevice, const char* filename )
{
    //  ڿ ȯѴ.
    release();

    //   .
    FILE *is;
    if( (is = fopen( filename, "rb" )) == NULL )
    {
        UPRINT("can't open %s", filename );
        return;
    }
    //  ĺڸ д´.
    size_t idLen = strlen( UMAP_FILE_ID );
    char *buffer = new char[ idLen ];

    size_t ret = fread( buffer, sizeof( char ), idLen, is );

    delete [] buffer;

    if( ret != idLen )
    {
        PRINTSCR("Error when reading a file id at %s", filename );
        goto LOAD_CLOSE_FILE;
    }

    //  ĺڸ ȮѴ.
    if( !strncmp( buffer, UMAP_FILE_ID, idLen ) )
    {
        PRINTSCR("invalid file : %s", filename );
        goto LOAD_CLOSE_FILE;
    }

    //  ũ⸦ д´
    int row, col;
    ret = fread( &row, sizeof( int ), 1, is );
    if( ret != 1 )
    {
        PRINTSCR("Error when reading the size of a row at %s", filename );
        goto LOAD_CLOSE_FILE;
    }
    fread( &col, sizeof( int ), 1, is );
    if( ret != 1 )
    {
        PRINTSCR("Error when reading the size of a column at %s", filename );
        goto LOAD_CLOSE_FILE;
    }
    if( row < 2 || col < 2 )
    {
        PRINTSCR("Invalid map size (%d, %d)", row, col);
        goto LOAD_CLOSE_FILE;
    }

    //   д´.
    int mapSize = row * col;
    float *heightData = new float[ mapSize ];
    ret = fread( heightData, sizeof( float ), mapSize, is );
    if( ret != mapSize )
    {
        PRINTSCR("Error when reading the height data of a map at %s", filename );
        goto LOAD_CLOSE_FILE;
    }

    //   Ѵ.
    create( pd3dDevice, row, col, heightData );

LOAD_CLOSE_FILE :
    // close a file
    fclose( is );
}

//
//  row-1,  col-1 ũ  ־ height  ؼ  Ѵ.
//
void UMap::create( LPDIRECT3DDEVICE9 pd3dDevice, int row, int col, float *pHeightMap )
{
    //  ڿ ȯѴ.
    release();

    m_pd3dDevice = pd3dDevice;
    m_row = row;
    m_col = col;
    m_pHeightMap = pHeightMap;

    // map texture Ѵ.
    if( FAILED(D3DXCreateTextureFromFile( pd3dDevice, "green096.jpg", &m_pMapTexture)) )
    {
        PRINTSCR("ؽ ε ");
        PostQuitMessage(0);
        return;
    }

    // vertex Ѵ.
    createVertices();
    // vertex normal Ѵ.
    calcVN();
    //  quad tree Ѵ.
    m_qtree.create( m_pd3dDevice, m_row, m_col, m_pMapVertices );

    m_bAvailable = true;
}

//
//  vertex Ѵ.
//
void UMap::createVertices()
{
    // vertex ޸ Ҵ
    m_pMapVertices = new MapVertex[ m_row*m_col ];
    ZeroMemory( m_pMapVertices, m_row*m_col*sizeof(MapVertex) );
    // vertex  
    int idx;
    for(int i=0; i<m_col; i++ )
    {
        for( int j=0; j<m_row; j++ )
        {
            idx = i*m_row+j;
            m_pMapVertices[idx].position = D3DXVECTOR3((float)j, m_pHeightMap[idx], (float)i);

            m_pMapVertices[idx].tu = (float)j;
            m_pMapVertices[idx].tv = (float)i;
        }
    }    
}

//
// vertex normal Ѵ.
// 
void UMap::calcVN()
{
    for( int i=0; i <m_col; i++ )
    {
        for( int j=0; j<m_row; j++ )
        {
            updateVN( j, i );
        }
    }
}

//
// face normal Ѵ.
//
#define FaceNormal(a,b,c,d,e,f) \
    p0 = &(m_pMapVertices[ (b)*m_row + (a) ].position); \
    p1 = &(m_pMapVertices[ (d)*m_row + (c) ].position); \
    p2 = &(m_pMapVertices[ (f)*m_row + (e) ].position); \
    D3DXPlaneFromPoints( &pl, p0, p1, p2 ); \
    D3DXPlaneNormalize( &pl, &pl );

//
// face normal Ѵ.
//
#define AddFaceNormal( e, f ) \
    (e).normal.x += (f).a; \
    (e).normal.y += (f).b; \
    (e).normal.z += (f).c; 

//
// vertex normal updateѴ.
//
// i+1 -----      vertex @ vertex normal vector 1,2,3,4 κ face normal ̴.     
//     |\  |\     1 face  i<col-1, j>0   Եȴ.
//     | 1 | \    2 face  i<col-1, j<row-1  Եȴ.
//     |  \| 2\   3 face  i>0, j>0  Եȴ.
//   i ----@----  4 face  i>0, j<row-1  Եȴ.
//      \  |\  |
//       \3| 4 |
//        \|  \|
// i-1     -----
//   j-1   j   j+1
//
inline void UMap::updateVN( int x, int y )
{
    int idx = y * m_row + x;
    D3DXVECTOR3 *p0, *p1, *p2;
    D3DXPLANE pl;

    // ʱȭ
    m_pMapVertices[ idx ].normal = D3DXVECTOR3( 0, 0, 0 );

    // 1 
    if( x > 0 && y < m_col-1)
    {
        FaceNormal( x, y, x-1, y+1, x, y+1 );
        AddFaceNormal( m_pMapVertices[ idx ], pl );

        FaceNormal( x, y, x-1, y, x-1, y+1 );
        AddFaceNormal( m_pMapVertices[ idx ], pl );
    }

    // 2 
    if( x < m_row-1 && y < m_col-1 )
    {
        FaceNormal( x, y, x, y+1, x+1, y );
        AddFaceNormal( m_pMapVertices[ idx ], pl );
    }

    // 3 
    if( x > 0 && y > 0 )
    {
        FaceNormal( x, y, x, y-1, x-1, y );
        AddFaceNormal( m_pMapVertices[ idx ], pl );
    }

    // 4 
    if( x < m_row-1 && y > 0 )
    {
        FaceNormal( x, y, x+1, y-1, x, y-1 );
        AddFaceNormal( m_pMapVertices[ idx ], pl );

        FaceNormal( x, y, x+1, y, x+1, y-1 );
        AddFaceNormal( m_pMapVertices[ idx ], pl );
    }

    // normalize
    D3DXVec3Normalize( &(m_pMapVertices[idx].normal), &(m_pMapVertices[idx].normal) );
}

//
// ڿ ȯѴ.
//
void UMap::release()
{
    // D3D Device ٸ  Ѵ.
    m_pd3dDevice = 0;

    // height map   ڿ
    Delete( m_pHeightMap );
    Delete( m_pMapVertices);

    // quad tree ڿ
    m_qtree.release();

    // texture Ѵ.
    ReleaseCom( m_pMapTexture );

    m_bAvailable = false;
}

//
// 콺 Ų  ãƳ.
//
bool UMap::findPickedPos( const D3DXVECTOR3 *vPickDir, const D3DXVECTOR3 *vPickOrg, D3DXVECTOR3 *vOut )
{
    return m_qtree.findPickedPos( vPickDir, vPickOrg, vOut );
}

//
//  frustum culling Ͽ renderingѴ.
//
void UMap::render( UFrustum &frustum )
{
    if( m_bAvailable )
    {
        // set texture
        m_pd3dDevice->SetTexture( 0, m_pMapTexture );
	    m_pd3dDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
	    m_pd3dDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	    m_pd3dDevice->SetSamplerState(0, D3DSAMP_MIPFILTER, D3DTEXF_POINT);

        // set wrap address mode
		m_pd3dDevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
		m_pd3dDevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);

        m_pd3dDevice->SetFVF( D3DFVF_MAPVERTEX );

        m_pd3dDevice->SetMaterial( &m_mtrlMap );

        m_qtree.render( frustum );
    }
}

//
// (x,z)  ̸ ˾Ƴ.
//
bool UMap::getHeight( const float x, float *y, const float z )
{
    if( x < 0.0f || x > m_row-1 || z < 0.0f || z > m_col-1 )
    {
        // unavailable region
        return false;
    }

    int row = (int)::floorf( x );
    int col = (int)::floorf( z );

    // C###D
    // ##  #
    // # # #
    // #  ##
    // A###B
    int idx = m_row*col+row;
    float A = m_pHeightMap[ idx ];
    float B = m_pHeightMap[ idx+1 ];
    float C = m_pHeightMap[ idx+m_row ];
    float D = m_pHeightMap[ idx+m_row+1 ];

    float dx = x - row;
    float dz = z - col;

    // ACB
    if( dz < 1.0f - dx )
    {
        // y = AB * dx + AC * dz
        *y = A + (B-A)*dx + (C-A)*dz;
    }
    // BCD
    else
    {
        // y = DC * (1-dx) + DB * (1-dz)
        *y = D + (C-D)*(1-dx) + (B-D) * (1-dz);
    }

    return true;
}

#ifdef EDITOR_BUILD
//
//  Ѵ.
//
void UMap::save( const char *filename )
{
    //    ƹ ͵  ʴ´.
    if( m_row == 0 )
    {
        UPRINT("Empty map");
        return;
    }
    //    .
    FILE *os;
    if( (os = fopen( filename, "wb" )) == NULL )
    {
        UPRINT("can't open %s", filename );
        return;
    }
    
    size_t ret;

    //   ĺ ڵ带 .
    size_t idLen = strlen( UMAP_FILE_ID );
    ret = fwrite( UMAP_FILE_ID, sizeof( char ), idLen, os );
    if( ret != idLen )
    {
        UPRINT("Error when writing a file id at %s", filename );
        goto SAVE_CLOSE_FILE;
    }

    //  ũ⸦ .
    //  ũ
    ret = fwrite( &m_row, sizeof( int ), 1, os );
    if( ret != 1 )
    {
        UPRINT("Error when writing the size of a row at %s", filename );
        goto SAVE_CLOSE_FILE;
    }
    //  ũ
    ret = fwrite( &m_col, sizeof( int ), 1, os );
    if( ret != 1 )
    {
        UPRINT("Error when writing the size of a column at %s", filename );
        goto SAVE_CLOSE_FILE;
    }

    //  Ÿ .
    int dataSize = m_row*m_col;
    ret = fwrite( m_pHeightMap, sizeof(float), dataSize, os );
    if( ret != dataSize )
    {
        UPRINT("Error when writing the height data of a map at %s", filename );
    }

SAVE_CLOSE_FILE :

    // close a file
    fclose( os );
}

//
// Ŀ  ̸ Ѵ.
//
void UMap::changeHeight( float delta )
{
    //    ׳ returnѴ.
    if( m_row == 0 )
    {
        return;
    }

    float oldYVal;
    int x, y;
    bool notXMin, notXMax, notYMin, notYMax;
    std::deque<int> vnUpdateList;
    std::deque<int>::iterator findResult;
    int newIdx;

    // ̸ Ѵ.
    for( std::deque<int>::iterator i = m_cursorArea.begin(); i != m_cursorArea.end(); i++ )
    {
        // vertex y  Ѵ.
        oldYVal = m_pHeightMap[ (*i) ];
        m_pHeightMap[ (*i) ] += delta;
        m_pMapVertices[ (*i) ].position.y += delta;

        x = (*i) % m_row;
        y = (*i) / m_row;

        // AABB update ʿ üũѴ.
        m_qtree.updateAABB( x, y, oldYVal, m_pHeightMap[(*i)] );

        // vertex normal  ɼ ִ  .
        notXMin = !( x == 0 );
        notXMax = !( x == m_row-1 );
        notYMin = !( y == 0 );
        notYMax = !( y == m_col-1 );
        
        // center
        newIdx = (*i);
        findResult = find( vnUpdateList.begin(), vnUpdateList.end(), newIdx );
        if( findResult == vnUpdateList.end() )
        {
            vnUpdateList.push_back( newIdx );
        }

        // west
        if( notXMin )
        {
            newIdx = (*i) - 1;
            findResult = find( vnUpdateList.begin(), vnUpdateList.end(), newIdx );
            if( findResult == vnUpdateList.end() )
            {
                vnUpdateList.push_back( newIdx );
            }
        }

        // north west
        if( notXMin && notYMax )
        {
            newIdx = (*i) + m_row - 1;
            findResult = find( vnUpdateList.begin(), vnUpdateList.end(), newIdx );
            if( findResult == vnUpdateList.end() )
            {
                vnUpdateList.push_back( newIdx );
            }
        }

        // north
        if( notYMax )
        {
            newIdx = (*i) + m_row;
            findResult = find( vnUpdateList.begin(), vnUpdateList.end(), newIdx );
            if( findResult == vnUpdateList.end() )
            {
                vnUpdateList.push_back( newIdx );
            }
        }

        // east
        if( notYMax )
        {
            newIdx = (*i) + 1;
            findResult = find( vnUpdateList.begin(), vnUpdateList.end(), newIdx );
            if( findResult == vnUpdateList.end() )
            {
                vnUpdateList.push_back( newIdx );
            }
        }

        // south east
        if( notYMin && notXMax )
        {
            newIdx = (*i) - m_row + 1;
            findResult = find( vnUpdateList.begin(), vnUpdateList.end(), newIdx );
            if( findResult == vnUpdateList.end() )
            {
                vnUpdateList.push_back( newIdx );
            }
        }

        // south
        if( notYMin )
        {
            newIdx = (*i) - m_row;
            findResult = find( vnUpdateList.begin(), vnUpdateList.end(), newIdx );
            if( findResult == vnUpdateList.end() )
            {
                vnUpdateList.push_back( newIdx );
            }
        }
    }

    // AABB  ȮѴ.
    m_qtree.commitAABBAndPatch();

    // vertex normal Ѵ.
    for( deque<int>::iterator i = vnUpdateList.begin(); i != vnUpdateList.end(); i++ )
    {
        x = (*i) % m_row;
        y = (*i) / m_row;
        updateVN( x, y );
    }
}

//
// 콺 ̿ ִ vertex ã´.
//   index returnϰ
//  -1 returnѴ.
//
int UMap::findCursorVertex( D3DXVECTOR3 *vPickDir, D3DXVECTOR3 *vPickOrg )
{
    return m_qtree.findPickedVertex( vPickDir, vPickOrg );
}

//
// Ŀ  Ѵ.
//
void UMap::makeCursorArea( int cursorIdx, int brushShape, int brushSize )
{
    // check if map is not initialized
    if( m_row == 0 )
    {
        return;
    }

    m_cursorArea.clear();

    if( cursorIdx == -1 )
    {
        return;
    }

    // find cursor position
    int map_x = cursorIdx % m_row;
    int map_y = cursorIdx / m_row;

    int left = map_x - brushSize + 1;
    if( left < 0 )
    {
        left = 0;
    }
    int right = map_x + brushSize;
    if( right > m_row )
    {
        right = m_row;
    }
    int top = map_y + brushSize;
    if( top > m_col )
    {
        top = m_col;
    }
    int bottom = map_y - brushSize + 1;
    if( bottom < 0 )
    {
        bottom = 0;
    }

    switch ( brushShape )
    {
    case MapEditor::CIRCLE_BRUSH :
        for( int i=bottom; i<top; i++ )
        {
            for( int j=left; j<right; j++ )
            {
                if( (brushSize-1)*(brushSize-1) >= (i - map_y)*(i - map_y) + (j - map_x)*(j - map_x ) )
                {
                    m_cursorArea.push_back( i*m_row + j );
                }
            }
        }
        break;
    case MapEditor::SQUARE_BRUSH :
        for( int i=bottom; i<top; i++ )
        {
            for( int j=left; j<right; j++ )
            {
                m_cursorArea.push_back( i*m_row +  j );
            }
        }
        break;
    }
}

#define setCursorVertex2( i, x ) \
    vCursor[i].position = (cur->position + x->position)/2; \
    vCursor[i].normal = x->normal;

#define setCursorVertex4( i, a, b, c ) \
    vCursor[i].position = (cur->position + a->position)/2; \
    vCursor[i].position.y = (b->position.y + c->position.y)/2; \
    vCursor[i].normal = a->normal;

//
// Ŀ  ׸.
//
void UMap::drawCursor( LPDIRECT3DDEVICE9 pd3dDevice, int cursorIdx, int brushShape, int brushSize )
{
    // map ϴ
    // cursor  ִ ȮѴ.
    if( cursorIdx == -1 || m_row == 0 )
    {
        return;
    }

    // world matrix Ѵ.
    D3DXMATRIX matWorld;
    pd3dDevice->GetTransform( D3DTS_WORLD, &matWorld );
    pd3dDevice->SetTransform( D3DTS_WORLD, &m_matCursor );

    // cursor material Ѵ.
    pd3dDevice->SetMaterial( &m_mtrlCursor );

    int cursorX, cursorY;
    MapVertex vCursor[9];
    MapVertex *nw, *n, *ne, *e, *se, *s, *sw, *w, *cur;
    bool xMin, xMax, yMin, yMax;
    for( deque<int>::iterator i = m_cursorArea.begin(); i != m_cursorArea.end() ; i++ )
    {
        // Ŀ  ǥ Ѵ.
        cursorX = (*i) % m_row;
        cursorY = (*i) / m_row;

        cur = &m_pMapVertices[ (*i) ];

        xMin = (cursorX == 0 );
        xMax = (cursorX == m_row - 1 );
        yMin = (cursorY == 0 );
        yMax = (cursorY == m_col - 1 );

        // check n
        if( !yMax )
        {
            n = &m_pMapVertices[ (*i) + m_row ];
            setCursorVertex2( 7, n );
        } else {
            vCursor[7] = *cur;
            vCursor[7].position.z += 0.5;
        }

        // check e
        if( !xMax )
        {
            e = &m_pMapVertices[ (*i) + 1];
            setCursorVertex2( 5, e );
        } else {
            vCursor[5] = *cur;
            vCursor[5].position.x += 0.5;
        }

        // check s
        if( !yMin )
        {
            s = &m_pMapVertices[ (*i) - m_row ];
            setCursorVertex2( 1, s );
        } else {
            vCursor[1] = *cur;
            vCursor[1].position.z -= 0.5;
        }

        // check w
        if( !xMin )
        {
            w = &m_pMapVertices[ (*i) - 1 ];
            setCursorVertex2( 3, w );
        } else {
            vCursor[3] = *cur;
            vCursor[3].position.x -= 0.5;
        }

        // check ne
        if( !(xMax | yMax) )
        {
            ne = &m_pMapVertices[ (*i) + m_row + 1 ];
            setCursorVertex4( 8, ne, n, e );
        } else {
            vCursor[8] = *cur;
            vCursor[8].position.x += 0.5;
            vCursor[8].position.z += 0.5;
        }

        // check se
        if( !(xMax | yMin) )
        {
            se = &m_pMapVertices[ (*i) - m_row + 1 ];

            setCursorVertex2( 2, se );
        } else {
            vCursor[2] = *cur;
            vCursor[2].position.x += 0.5;
            vCursor[2].position.z -= 0.5;
        }

        // check sw
        if( !(xMin | yMin) )
        {
            sw = &m_pMapVertices[ (*i) - m_row - 1 ];
            setCursorVertex4( 0, sw, s, w );
        } else {
            vCursor[0] = *cur;
            vCursor[0].position.x -= 0.5;
            vCursor[0].position.z -= 0.5;
        }

        // check nw
        if( !(xMin | yMax) )
        {
            nw = &m_pMapVertices[ (*i) + m_row - 1 ];
            setCursorVertex2( 6, nw );
        } else {
            vCursor[6] = *cur;
            vCursor[6].position.x -= 0.5;
            vCursor[6].position.z += 0.5;
        }

        // cur
        vCursor[4] = *cur;

        // ׸.
        pd3dDevice->DrawIndexedPrimitiveUP(
            D3DPT_TRIANGLELIST,
            0,
            9,
            8,
            pCursorIndex,
            D3DFMT_INDEX16,
            vCursor,
            sizeof( MapVertex ) );
    }
    // world matrix ǵ.
    pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
}

#endif