#include "UTimer.h"

//
// initialize
//
// if the hi-res timer is present, the tick rate is stored. Otherwise,
// the timer should be use GetTickCount()
//
void UTimer::init()
{
    LARGE_INTEGER ticksPerSecond;
    if( !QueryPerformanceFrequency( &ticksPerSecond ) )
    {
        // system doesn't support hi-res timer
        m_timer = new NormalTimer();
    }
    else
    {
        m_timer = new HPTimer();
    }
}

//
// release
//
void UTimer::release()
{
    if( m_timer != NULL )
    {
        delete m_timer;
    }
}

//
// get FPS( frames per second )
//
int UTimer::getFPS()
{
    if( m_timer != NULL )
    {
        return m_timer->getFPS();
    }
    return 0;
}

//
// get FPS( virtual function )
//
int _Timer::getFPS()
{
    return 0;
}

//
// hi-res timer constructor
//
HPTimer::HPTimer()
{
    QueryPerformanceFrequency( &m_ticksPerSecond );
    QueryPerformanceCounter( &m_startTime );
    m_FPS = 0;
    m_lastTime = 0;
}

//
// get FPS
//
// if elapsed time is not over 1 sec 0 is returned
//
int HPTimer::getFPS()
{
    LARGE_INTEGER getCurrentTime;
    QueryPerformanceCounter( &getCurrentTime );
    float currentTime = (float)getCurrentTime.QuadPart / (float)m_ticksPerSecond.QuadPart;
    ++m_FPS;
    if( currentTime - m_lastTime > 1.0f )
    {
        m_lastTime = currentTime;
        
        int ret = m_FPS;
        m_FPS = 0;
        return ret;
    }
    return 0;
}

//
// normal timer construtor
//
NormalTimer::NormalTimer()
{
    m_startTime = GetTickCount() * 0.001f;
    m_FPS = 0;
    m_lastTime = 0;
}

//
// get FPS
//
int NormalTimer::getFPS()
{
    float currentTime = GetTickCount() * 0.001f;
    ++m_FPS;

    if( currentTime - m_lastTime > 1.0f )
    {
        m_lastTime = currentTime;
        int ret = m_FPS;
        m_FPS = 0;

        return ret;
    }

    return 0;
}