//
// MapEditor2
//

#define _WIN32_WINNT 0x0400

#include <windows.h>
#include <windowsx.h>
#include "MapEditor.h"
#include "resource.h"
#include "udebug.h"

//
//  
//
MapEditor g_me;    // MapEditor Instance
HWND g_toolDlg;    // tool dialog
HWND g_minimapDlg; // minimap dialog

struct MinimapInfo
{
    D3DXVECTOR3 edgeVertices[4];
    float xMin;
    float xMax;
    float zMin;
    float zMax;
    RECT rect;
} g_minimapInfo;

//
//  ũ   structure
//
struct MapSize
{
    int row;
    int col;
};

//
// ʱȭ
//
HRESULT init( HWND hWnd )
{
    return g_me.init( hWnd );
}

//
// message handler for a dialog to input map size
//
BOOL CALLBACK SizeInputDlgProc (
    HWND    hDlg, 
    UINT    message, 
    WPARAM  wParam, 
    LPARAM  lParam ) 
{
    // for receiveing parameters
    static MapSize msd, *pmsd;
    char buffer [6];

    switch ( message ) {
    case WM_INITDIALOG :
        // set variable for sending the size of a map
        pmsd = (MapSize *)lParam;
        msd = *pmsd;

        // set the limit of edit controls to 3
        SendDlgItemMessage( hDlg, IDC_EDIT_ROW, EM_LIMITTEXT, 3, NULL );
        SendDlgItemMessage( hDlg, IDC_EDIT_COL, EM_LIMITTEXT, 3, NULL );

        return TRUE;
    case WM_COMMAND :
        switch (LOWORD(wParam)) {
        case IDOK :
            // get row
            if( GetDlgItemText( hDlg, IDC_EDIT_ROW, buffer, 6) == 0 ) {
                MessageBox( NULL, "please INPUT row size", NULL, MB_OK );
                return FALSE;
            }
            msd.row = atoi( buffer );
            if( msd.row > UMap::MAP_MAX_SIZE )
				msd.row = UMap::MAP_MAX_SIZE;

			// get col
            if( GetDlgItemText( hDlg, IDC_EDIT_COL, buffer, 6) == 0 ) {
                MessageBox( NULL, "please INPUT column size", NULL, MB_OK );
                return FALSE;
            }
            msd.col = atoi( buffer );
			if( msd.col > UMap::MAP_MAX_SIZE )
				msd.col = UMap::MAP_MAX_SIZE;
            
            // send the size of a map to parent
            *pmsd = msd;
            // close dialog
            EndDialog( hDlg, TRUE );
            return TRUE;
        case IDCANCEL :
            // lcose dialog
            EndDialog( hDlg, FALSE );
            return TRUE;
        }
    }
    return FALSE;
}

// 
//   
//
void createNewMap( HWND hWnd )
{
    MapSize ms;
	if( DialogBoxParam( GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_INPUTMAPSIZE), hWnd, (DLGPROC)SizeInputDlgProc, (LPARAM)&ms ) == IDOK) {
        g_me.createNewMap( ms.row, ms.col );
        g_minimapInfo.xMin = 0;
        g_minimapInfo.xMax = (float)ms.row;
        g_minimapInfo.zMin = 0;
        g_minimapInfo.zMax = (float)ms.col;
        InvalidateRect( g_minimapDlg, NULL, TRUE );
	}
}

//
// save to a file
//
void saveMap( HWND hWnd )
{
    char fname[1024];
    memset( fname, 0, sizeof(char)*1024 );
    OPENFILENAME ofn;

    memset(&ofn, 0, sizeof(OPENFILENAME));
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = hWnd;
    ofn.lpstrFilter = "map file (*.map)\0*.map\0All\0*.*\0";
    ofn.lpstrDefExt = "map";
    ofn.lpstrFile = fname;
    ofn.nMaxFile = 255;
    ofn.lpstrFileTitle = "Save Map File";
    ofn.Flags = OFN_OVERWRITEPROMPT;

    if (GetSaveFileName(&ofn)==TRUE)
    {
        g_me.save( fname );
    }
}

//
// open to a file
//
void loadMap( HWND hWnd )
{
    char fname[1024];
    memset( fname, 0, sizeof(char)*1024 );

    OPENFILENAME ofn;
    memset(&ofn, 0, sizeof(OPENFILENAME));
    ofn.lStructSize = sizeof(OPENFILENAME);
    ofn.hwndOwner = hWnd;
    ofn.lpstrFilter = "map file (*.map)\0*.map\0All\0*.*\0";
    ofn.lpstrDefExt = "map";
    ofn.lpstrFile = fname;
    ofn.nMaxFile = 255;
    ofn.lpstrFileTitle = "Load Map File";
    ofn.Flags = OFN_OVERWRITEPROMPT;

    if (GetOpenFileName(&ofn)==TRUE)
    {
        g_me.load( fname );
        int row, col;
        g_me.getMapSize( &row, &col );
        g_minimapInfo.xMin = 0;
        g_minimapInfo.xMax = (float)row;
        g_minimapInfo.zMin = 0;
        g_minimapInfo.zMax = (float)col;
        InvalidateRect( g_minimapDlg, NULL, TRUE );
    }
}

//
// ڿ 
//
void cleanUp()
{
    // destroy modaless dialogs
    DestroyWindow( g_minimapDlg );
    DestroyWindow( g_toolDlg );
    // release resources of a map editor
    g_me.release();
}

//
// view  
//
void nextViewMode()
{
    g_me.nextViewMode();
}

//
// render the scene
//
inline void render()
{
    g_me.render();
}

//
// do command
//
inline void doCmd( int x, int y )
{
    g_me.doCmd( x, y );
}

//
// update cursor position
//
inline void updateCursorPos( int x, int y )
{
    g_me.updateCursorPos( x, y );
}

//
// message handler
//
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    // store old mouse position
    static int oldx = 0, oldy = 0;
    switch( msg )
    {
    case WM_DESTROY:
        cleanUp();
        PostQuitMessage( 0 );
        return 0;
        
    case WM_COMMAND:
        switch( LOWORD( wParam ) )
        {
        case IDM_FILE_NEW :
        case ID_ACC_NEWMAP :
            // create a new map
            createNewMap( hWnd );
            break;
        case IDM_FILE_OPEN :
        case ID_ACC_OPENMAP :
            // open a map
            loadMap( hWnd );
            break;
        case IDM_FILE_SAVE :
        case ID_ACC_SAVEMAP :
            // save a map
            saveMap( hWnd );
            break;
        case IDM_VIEW_TOOL :
            // show tool dialog
            ShowWindow( g_toolDlg, SW_SHOWDEFAULT );
            break;
        case IDM_VIEW_MINIMAP :
            // show minimap dialog
            ShowWindow( g_minimapDlg, SW_SHOWDEFAULT );
            break;
        case IDM_FILE_EXIT :
            // terminate program
            cleanUp();
            PostQuitMessage( 0 );
            return 0;
        default :
            break;
        }
        break;
    case WM_RBUTTONDOWN :
    case WM_LBUTTONDOWN :
    case WM_MOUSEMOVE :
        {
            // set focus to this
            SetFocus( hWnd );
            // get the position of a cursor
            int x = GET_X_LPARAM( lParam );
            int y = GET_Y_LPARAM( lParam );
            // update the position of a cursor on a map
            updateCursorPos( x, y );
            // if left button is pushed do some command
            if( wParam == MK_LBUTTON )
            {
                doCmd( x, y );
            }
            // if right button is pushed change the position of camera
            else if( wParam == MK_RBUTTON )
            {
                // calculate the difference between new and old position of mouse
                if( oldx != 0 && oldy != 0 )
                {
                    // rotate camera
                    RECT cRect;
                    GetClientRect( hWnd, &cRect );
                    g_me.rotateCamera( (float)(oldx - x)/(cRect.right-cRect.left), (float)(oldy - y)/(cRect.bottom-cRect.top) );
                    // update minimap window
                    InvalidateRect( g_minimapDlg, NULL, TRUE );
                    
                }
                // store old position of mouse
                oldx = x;
                oldy = y;
            }
        }
        break;
    case WM_RBUTTONUP :
        oldx = 0;
        oldy = 0;
        break;
    case WM_MOUSEWHEEL :
        {
            // get wheel delta
            short zDelta = GET_WHEEL_DELTA_WPARAM(wParam);
            float delta = (float)(zDelta / WHEEL_DELTA);
            // zoom a camera
            g_me.zoomCamera( delta );
            // update a frustum of minimap
            InvalidateRect( g_minimapDlg, NULL, TRUE );
        }
        break;
    case WM_CHAR :
        {
            char c = (TCHAR) wParam;
            switch( c )
            {
            case 'a' :
                g_me.moveCamera( MapEditor::LEFT );
                break;
            case 'd' :
                g_me.moveCamera( MapEditor::RIGHT );
                break;
            case 'w' :
                g_me.moveCamera( MapEditor::FRONT );
                break;
            case 's' :
                g_me.moveCamera( MapEditor::BACK );
                break;
            }
        }
        break;
    case WM_EXITSIZEMOVE:
/*        {
        RECT rcWindowClient;
        GetClientRect( hWnd, &rcWindowClient );
        g_me.resetWindowSize( rcWindowClient.right - rcWindowClient.right, rcWindowClient.bottom - rcWindowClient.top );
        }*/
        break;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}

//
// message handler of tool dialog
//
int CALLBACK toolDlgMsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
    case WM_INITDIALOG :
        {
        // get
        char *cmd = g_me.getCmd();
        char *brushSize = g_me.getBrushSize();
        char *brushShape = g_me.getBrushShape();
        char buffer[20];
        // set command string
        ZeroMemory( buffer, sizeof(buffer) );
        sprintf( buffer, "Apply Height : %s", cmd );
        SetDlgItemText( hWnd, IDC_CHECK_APPLYHEIGHT, buffer );
        // set brush size
        ZeroMemory( buffer, sizeof(buffer) );
        sprintf( buffer, "Size : %s", brushSize );
        SetDlgItemText( hWnd, IDC_STATIC_SIZE, buffer );
        // set brush shape
        ZeroMemory( buffer, sizeof(buffer) );
        sprintf( buffer, "Shape : %s", brushShape );
        SetDlgItemText( hWnd, IDC_STATIC_SHAPE, buffer );
        return TRUE;
        }
    case WM_COMMAND:
        switch( LOWORD( wParam ) )
        {
        case IDCANCEL : // push terminate mark
            ShowWindow( hWnd, SW_HIDE );
            return TRUE;
        // tool command
        case IDC_BUTTON_RAISE :
            g_me.changeCmd( MapEditor::CMD_RAISE );
            SetDlgItemText( g_toolDlg, IDC_CHECK_APPLYHEIGHT, "Apply Height : Raise" );
            return TRUE;
        case IDC_BUTTON_LOWER :
            g_me.changeCmd( MapEditor::CMD_LOWER );
            SetDlgItemText( g_toolDlg, IDC_CHECK_APPLYHEIGHT, "Apply Height : Lower" );
            return TRUE;
        case IDC_BUTTON_SIZE1 :
            g_me.changeBrushSize( MapEditor::BRUSH_1 );
            SetDlgItemText( g_toolDlg, IDC_STATIC_SIZE, "Size : 1" );
            return TRUE;
        case IDC_BUTTON_SIZE2 :
            g_me.changeBrushSize( MapEditor::BRUSH_2 );
            SetDlgItemText( g_toolDlg, IDC_STATIC_SIZE, "Size : 2" );
            return TRUE;
        case IDC_BUTTON_SIZE3 :
            g_me.changeBrushSize( MapEditor::BRUSH_3 );
            SetDlgItemText( g_toolDlg, IDC_STATIC_SIZE, "Size : 3" );
            return TRUE;
        case IDC_BUTTON_SIZE5 :
            g_me.changeBrushSize( MapEditor::BRUSH_5 );
            SetDlgItemText( g_toolDlg, IDC_STATIC_SIZE, "Size : 5" );
            return TRUE;
        case IDC_BUTTON_SIZE8 :
            g_me.changeBrushSize( MapEditor::BRUSH_8 );
            SetDlgItemText( g_toolDlg, IDC_STATIC_SIZE, "Size : 8" );
            return TRUE;
        case IDC_BUTTON_CIRCLE :
            g_me.changeBrushShape( MapEditor::CIRCLE_BRUSH );
            SetDlgItemText( g_toolDlg, IDC_STATIC_SHAPE, "Shape : Circle" );
            return TRUE;
        case IDC_BUTTON_SQUARE :
            g_me.changeBrushShape( MapEditor::SQUARE_BRUSH );
            SetDlgItemText( g_toolDlg, IDC_STATIC_SHAPE, "Shape : Square" );
            return TRUE;
        }
        break;
    }
    return FALSE;
}

#define TransDXtoWindow( a ) \
    diff = edgeVertices[a].x - g_minimapInfo.xMin; \
    frustum[a].x = g_minimapInfo.rect.left + (long)((diff < 0 )?0:((diff>areaWidth)?rectWidth: (diff/areaWidth*rectWidth)));\
    diff = edgeVertices[a].z - g_minimapInfo.zMin; \
    frustum[a].y = g_minimapInfo.rect.bottom - (long)(( diff < 0 )?0:((diff>areaHeight)?rectHeight: (diff/areaHeight*rectHeight)));

//
// translate DX coordinate to Windows coordinate
//
void translateDXtoWindows( const D3DXVECTOR3 *edgeVertices, POINT *frustum )
{
    float rectWidth = (float)(g_minimapInfo.rect.right-g_minimapInfo.rect.left);
    float rectHeight = (float)(g_minimapInfo.rect.bottom-g_minimapInfo.rect.top);
    float areaWidth = g_minimapInfo.xMax - g_minimapInfo.xMin;
    float areaHeight = g_minimapInfo.zMax - g_minimapInfo.zMin;
    float diff;

    TransDXtoWindow( 0 );
    TransDXtoWindow( 1 );
    TransDXtoWindow( 2 );
    TransDXtoWindow( 3 );
}

//
// translate Windows to DX
//
void translateWindowsToDX( int x, int y, D3DXVECTOR3 *dx )
{
    float rectWidth = (float)(g_minimapInfo.rect.right-g_minimapInfo.rect.left);
    float rectHeight = (float)(g_minimapInfo.rect.bottom-g_minimapInfo.rect.top);
    float areaWidth = g_minimapInfo.xMax - g_minimapInfo.xMin;
    float areaHeight = g_minimapInfo.zMax - g_minimapInfo.zMin;

    dx->x = g_minimapInfo.xMin + (x) / rectWidth * areaWidth;
    dx->y = 0;
    dx->z = g_minimapInfo.zMax - (y) / rectHeight * areaHeight;
}

//
// message handler of minimap dialog
//
int WINAPI minimapDlgMsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
    case WM_PAINT :
        if( g_me.isMapAvailable )
        {
            PAINTSTRUCT ps;
            BeginPaint( hWnd, &ps );

            // draw the total area of a minimap
            SelectObject( ps.hdc, GetStockObject( BLACK_BRUSH ) );

            GetClientRect( hWnd, &(g_minimapInfo.rect) );
            Rectangle( ps.hdc, g_minimapInfo.rect.left, g_minimapInfo.rect.top, g_minimapInfo.rect.right , g_minimapInfo.rect.bottom );

            // draw the view-frustum
            static HPEN yellowPen = CreatePen( PS_SOLID, 0, RGB( 255, 255, 0 ) );

            g_me.getFrustumOnXZ( g_minimapInfo.edgeVertices );
            SelectObject( ps.hdc, yellowPen );
            POINT frustum [4];
            translateDXtoWindows( g_minimapInfo.edgeVertices, frustum );
            Polygon( ps.hdc, frustum, 4 );

            EndPaint( hWnd, &ps );

            return TRUE;
        }
        break;
    case WM_COMMAND :
        switch( LOWORD( wParam ) )
        {
        case IDCANCEL : // push terminate mark
            ShowWindow( hWnd, SW_HIDE );
            return TRUE;
        }
        break;
    case WM_LBUTTONDOWN :
    case WM_MOUSEMOVE :
        if( g_me.isMapAvailable && wParam == MK_LBUTTON )
        {
            int x = GET_X_LPARAM( lParam );
            int z = GET_Y_LPARAM( lParam );
            D3DXVECTOR3 lookat;
            translateWindowsToDX( x, z, &lookat );

            g_me.moveCamera( &lookat );

            //UPRINT("(%d,%d) to (%f, %f, %f)", x, z, eye.x, eye.y, eye.z );

            InvalidateRect( hWnd, NULL, TRUE );
            return TRUE;
        }

    }
    return FALSE;
}

//
// application's entry point
//
int WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    // register window's class
    WNDCLASSEX wc = {
        sizeof( WNDCLASSEX ),   // cbSize
        CS_CLASSDC,             // style
        MsgProc,                // lpfnWndProc
        0L,                     // cbClassExtra
        0L,                     // cbWndExtra
        GetModuleHandle( NULL ),// hInstance
        NULL,                   // hIcon
        NULL,                   // hCursor
        NULL,                   // hbrBackground
        MAKEINTRESOURCE( IDR_MENU ), //NULL,                   // lpszMenuName
        "MapEditor",            // lpszClassName
        NULL                    // hIconSm        
    };
    RegisterClassEx( &wc );

    // create the application's window
    HWND hWnd = CreateWindow( "MapEditor", // lpClassName
                              "Map Editor", // lpWindowName
                              WS_OVERLAPPED | WS_SYSMENU, // dwStyle
                              100, // x
                              100, // y
                              1024, // nWidth
                              768, // nHeight
                              GetDesktopWindow(), // hWndParent
                              NULL, // hMenu
                              wc.hInstance, // hInstance
                              NULL); // lpParam
    HACCEL haccel; // handle to accellerator table
    // Load the accelerator table
    haccel = LoadAccelerators( GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_ACC_MAPEDITOR));

    // Initialize application
    if( SUCCEEDED(init( hWnd )) )
    {
        // Show the window
        ShowWindow( hWnd, SW_SHOWDEFAULT );
        UpdateWindow( hWnd );

        RECT parentRect, rect, rect2;
        GetWindowRect( hWnd, &parentRect );
        GetClientRect( hWnd, &rect );
        parentRect.top = parentRect.bottom - rect.bottom;

        // Create Minimap Dialog
        g_minimapDlg = CreateDialog( hInst, MAKEINTRESOURCE( IDD_MINIMAP ), hWnd, minimapDlgMsgProc );
        ShowWindow( g_minimapDlg, SW_SHOWDEFAULT );
        GetWindowRect( g_minimapDlg, &rect );
        MoveWindow( g_minimapDlg, parentRect.right - (rect.right - rect.left ) - 10, parentRect.top , rect.right - rect.left, rect.bottom - rect.top, TRUE );

        // Create Tool Dialog
        g_toolDlg = CreateDialog( hInst, MAKEINTRESOURCE( IDD_TOOL ), hWnd, toolDlgMsgProc );
        ShowWindow( g_toolDlg, SW_SHOWDEFAULT );
        GetWindowRect( g_toolDlg, &rect2 );
        MoveWindow( g_toolDlg, parentRect.right - (rect2.right - rect2.left )  - 10, parentRect.top + rect.bottom - rect.top , rect2.right - rect2.left, rect2.bottom - rect2.top, TRUE );

        // Enter the message loop
        MSG msg;
        ZeroMemory( &msg, sizeof( msg ) );
        while( msg.message != WM_QUIT )
        {
            if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
            {
                if( !TranslateAccelerator( hWnd,   // handle to receiving window
                                           haccel, // handle to active accelerator table
                                           &msg) )
                {
                    TranslateMessage( &msg );
                    DispatchMessage( &msg );
                }
            }
            else
            {
                render();
            }
        }
    }
    UnregisterClass( "MapEditor", wc.hInstance );
    return 0;
}