#include <stdarg.h>
#include <stdio.h>
#include <windows.h>
#include "udebug.h"

namespace ulog {
bool g_logInit = false;
}

//
// initialize a log module
//
void initLog( LPDIRECT3DDEVICE9 pd3dDevice, RECT region, size_t maxMsgLines )
{
    Log::init( pd3dDevice, region, maxMsgLines );
    ulog::g_logInit = true;
}

//
// release a log module
//
void releaseLog()
{
    Log::release();
    ulog::g_logInit = false;
}

//
// α  ϸ α׿ ׷  MessageBox Ͽ α׸ Ѵ.
//
void UPRINT( char *format, ... )
{ 
    char buffer [256];
    va_list argptr;

    va_start( argptr, format );
    vsprintf( buffer, format, argptr );
    va_end( argptr );

    if( ulog::g_logInit ) {
        Log::print( buffer );
    } else {
        MessageBox( 0, buffer, 0, 0 );
    }
}

//
// MessageBox Ͽ α׸ Ѵ.
//
void PRINTSCR( char *format, ... )
{
    char buffer [256];
    va_list argptr;

    va_start( argptr, format );
    vsprintf( buffer, format, argptr );
    va_end( argptr );

    MessageBox( 0, buffer, 0, 0 );
}

// declare static members
LPD3DXSPRITE Log::m_pTextSprite = NULL;
LPD3DXFONT Log::m_pFont = NULL;
deque<string> Log::m_messages;
RECT Log::m_region;
size_t Log::MAX_MSG_LINES;
int Log::m_fontSize;
char *Log::m_strFPS;

//
// initialize log module
//
void Log::init( LPDIRECT3DDEVICE9 pd3dDevice, RECT region, size_t maxMsgLines )
{
    HRESULT hr;
    CopyRect( &m_region, &region );
    MAX_MSG_LINES = maxMsgLines;
    m_fontSize = 0;

    // create a text sprite
    hr = D3DXCreateSprite( pd3dDevice, &m_pTextSprite );

    // create a font
    HDC hDC = GetDC( NULL );
    int nLogPixelsY = GetDeviceCaps(hDC, LOGPIXELSY);
    ReleaseDC( NULL, hDC );

    int nHeight = -m_fontSize * nLogPixelsY / 72;
    hr = D3DXCreateFont( pd3dDevice,            // D3D device
                         nHeight,               // Height
                         0,                     // Width
                         FW_NORMAL,               // Weight
                         1,                     // MipLevels, 0 = autogen mipmaps
                         FALSE,                 // Italic
                         DEFAULT_CHARSET,       // CharSet
                         OUT_DEFAULT_PRECIS,    // OutputPrecision
                         DEFAULT_QUALITY,       // Quality
                         DEFAULT_PITCH | FF_DONTCARE, // PitchAndFamily
                         "FixedSys",              // pFaceName
                         &m_pFont);              // ppFont
}

//
// release resources
//
void Log::release()
{
    if( m_pTextSprite != NULL )
    {
        m_pTextSprite->Release();
        m_pTextSprite = NULL;
    }
    if( m_pFont != NULL )
    {
        m_pFont->Release();
        m_pFont = NULL;
    }
    m_messages.clear();
}

//
// print a message
//
void Log::print( string msg )
{
    m_messages.push_back( msg );
    if( m_messages.size() > MAX_MSG_LINES )
    {
        m_messages.erase( m_messages.begin() );
    }
}

void Log::print( char *buffer )
{
    print( string(buffer ) );
}

//
// render a message
//
void Log::render( LPDIRECT3DDEVICE9 pd3dDevice )
{
    int count = 0;
    RECT rc;

    m_pTextSprite->Begin( D3DXSPRITE_ALPHABLEND | D3DXSPRITE_SORT_TEXTURE );

    // draw FPS
    if( m_strFPS != NULL )
    {
        SetRect( &rc, m_region.left, m_region.top + 15*(count), 0, 0 );
        m_pFont->DrawText( m_pTextSprite, m_strFPS, -1, &rc, DT_NOCLIP, D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
    }
    count++;

    // draw log messages
    if( m_messages.size() != 0 )
    {
        for( deque<string>::iterator i = m_messages.begin(); i != m_messages.end(); i++ )
        {
            SetRect( &rc, m_region.left, m_region.top + 15*(count), 0, 0 );
            m_pFont->DrawText( m_pTextSprite, (*i).c_str(), -1, &rc, DT_NOCLIP, D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
            count ++;
        }
    }
    m_pTextSprite->End();
}