#include "util.h"

//
//   ̷ ﰢ face normal Ѵ.
// 
// -   ð  Էµȴٰ Ѵ.
// - (P2-P0) x (P1-P0)
void calcFaceNormal( D3DXVECTOR3 p[], D3DXVECTOR3 *out)
{
    // (p2 - p0 )
    D3DXVECTOR3 v1 = p[2] - p[0];
    // (p1 - p0 )
    D3DXVECTOR3 v2 = p[1] - p[0];
    // v1 x v2
    D3DXVec3Cross( out, &v1, &v2 );
    // normalize
    D3DXVec3Normalize( out, out );
}

//
// calculate view direction and origin
//
// vDir     [out] pick direction
// vOrig    [out] pick origin
// x        [in]  x
// y        [in]  y
// matView  [in]  view matrix
// matProj  [in]  projection matrix
// viewport [in]  viewport
//
void calcViewDir( D3DXVECTOR3 *vPickRayDir, D3DXVECTOR3 *vPickRayOrig, int x, int y, const D3DXMATRIX *matView, const D3DXMATRIX *matProj, const D3DVIEWPORT9 *vp )
{
    D3DXVECTOR3 v;
    v.x = (( (( x - vp->X )*2.0f/(vp->Width))  -1.0f) - matProj->_31 )/matProj->_11;
    v.y = ((-(( y - vp->Y )*2.0f/(vp->Height)) -1.0f) - matProj->_32 )/matProj->_22;
    v.z = 1.0f;

    D3DXMATRIX m;
    D3DXMatrixInverse( &m, NULL, matView );
    vPickRayDir->x = v.x*m._11 + v.y*m._21 + v.z*m._31;
    vPickRayDir->y = v.x*m._12 + v.y*m._22 + v.z*m._32;
    vPickRayDir->z = v.x*m._13 + v.y*m._23 + v.z*m._33;
    vPickRayOrig->x = m._41;
    vPickRayOrig->y = m._42;
    vPickRayOrig->z = m._43;
}

//-----------------------------------------------------------------------------
// Name: intersectTriangle()
// Desc: Given a ray origin (orig) and direction (dir), and three vertices of
//       of a triangle, this function returns TRUE and the interpolated texture
//       coordinates if the ray intersects the triangle
//-----------------------------------------------------------------------------
bool intersectTriangle( const D3DXVECTOR3& v0,
                        const D3DXVECTOR3& v1,
                        const D3DXVECTOR3& v2,
                        const D3DXVECTOR3& orig,
                        const D3DXVECTOR3& dir,
                        FLOAT* t, FLOAT* u, FLOAT* v )
{
	// Find vectors for two edges sharing vert0
    D3DXVECTOR3 edge1 = v1 - v0;
    D3DXVECTOR3 edge2 = v2 - v0;

    // Begin calculating determinant - also used to calculate U parameter
    D3DXVECTOR3 pvec;
    D3DXVec3Cross( &pvec, &dir, &edge2 );

    // If determinant is near zero, ray lies in plane of triangle
    FLOAT det = D3DXVec3Dot( &edge1, &pvec );

    D3DXVECTOR3 tvec;
    if( det > 0 )
    {
        tvec = orig - v0;
    }
    else
    {
        tvec = v0 - orig;
        det = -det;
    }

    if( det < 0.0001f )
        return FALSE;

    // Calculate U parameter and test bounds
    *u = D3DXVec3Dot( &tvec, &pvec );
    if( *u < 0.0f || *u > det )
        return FALSE;

    // Prepare to test V parameter
    D3DXVECTOR3 qvec;
    D3DXVec3Cross( &qvec, &tvec, &edge1 );

    // Calculate V parameter and test bounds
    *v = D3DXVec3Dot( &dir, &qvec );
    if( *v < 0.0f || *u + *v > det )
        return FALSE;

    // Calculate t, scale parameters, ray intersects triangle
    *t = D3DXVec3Dot( &edge2, &qvec );
    FLOAT fInvDet = 1.0f / det;
    *t *= fInvDet;
    *u *= fInvDet;
    *v *= fInvDet;

    return TRUE;
}